#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_fit.h>
#include "fit.h"

int expd3_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;
  
  double A1=gsl_vector_get(x,0);
  double t1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double t2=gsl_vector_get(x,3);
  double A3=gsl_vector_get(x,4);
  double t3=gsl_vector_get(x,5);
  double y0=gsl_vector_get(x,6);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A1 * exp (-X[i]*t1) + A2 * exp (-X[i]*t2) + A3 * exp (-X[i]*t3) +y0;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int expd3_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A1=gsl_vector_get(x,0);
  double l1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double l2=gsl_vector_get(x,3);
  double A3=gsl_vector_get(x,4);
  double l3=gsl_vector_get(x,5);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) +y0  */
      /* and the xj are the parameters (A1,l1,A2,l2,y0) */
      double t = X[i];
      double s = sigma[i];
      double e1 = exp(-t*l1);
	  double e2 = exp(-t*l2);
	  double e3 = exp(-t*l3);
		
      gsl_matrix_set (J, i, 0, e1/s);
      gsl_matrix_set (J, i, 1, -t * A1 * e1/s);
      gsl_matrix_set (J, i, 2, e2/s);
	  gsl_matrix_set (J, i, 3, -t * A2 * e2/s);
	  gsl_matrix_set (J, i, 4, e3/s);
	  gsl_matrix_set (J, i, 5, -t * A3 * e3/s);
      gsl_matrix_set (J, i, 6, 1/s);
    }
  return GSL_SUCCESS;
}

int expd3_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expd3_f (x, params, f);
  expd3_df (x, params, J);

  return GSL_SUCCESS;
}

int expd2_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;
  
  double A1=gsl_vector_get(x,0);
  double t1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double t2=gsl_vector_get(x,3);
  double y0=gsl_vector_get(x,4);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A1 * exp (-X[i]*t1) + A2 * exp (-X[i]*t2) + y0;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int expd2_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A1=gsl_vector_get(x,0);
  double l1=gsl_vector_get(x,1);
  double A2=gsl_vector_get(x,2);
  double l2=gsl_vector_get(x,3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A1 * exp(-xi*l1) + A2 * exp(-xi*l2) +y0  */
      /* and the xj are the parameters (A1,l1,A2,l2,y0) */
      double t = X[i];
      double s = sigma[i];
      double e1 = exp(-t*l1);
	  double e2 = exp(-t*l2);
		
      gsl_matrix_set (J, i, 0, e1/s);
      gsl_matrix_set (J, i, 1, -t * A1 * e1/s);
      gsl_matrix_set (J, i, 2, e2/s);
	  gsl_matrix_set (J, i, 3, -t * A2 * e2/s);
      gsl_matrix_set (J, i, 4, 1/s);
    }
  return GSL_SUCCESS;
}

int expd2_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expd2_f (x, params, f);
  expd2_df (x, params, J);

  return GSL_SUCCESS;
}

int expb_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  double b = gsl_vector_get (x, 2);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A * exp (-lambda * X[i]) + b;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int expb_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A * exp(-lambda * i) + b  */
      /* and the xj are the parameters (A,lambda,b) */
      double t = X[i];
      double s = sigma[i];
      double e = exp(-lambda * t);
      gsl_matrix_set (J, i, 0, e/s);
      gsl_matrix_set (J, i, 1, -t * A * e/s);
      gsl_matrix_set (J, i, 2, 1/s);

    }
  return GSL_SUCCESS;
}

int expb_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expb_f (x, params, f);
  expb_df (x, params, J);

  return GSL_SUCCESS;
}

int expg_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);
  double b = gsl_vector_get (x, 2);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A * exp (lambda * X[i]) + b;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int expg_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 0);
  double lambda = gsl_vector_get (x, 1);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /*       Yi = A * exp(lambda * i) + b  */
      /* and the xj are the parameters (A,lambda,b) */
      double t = X[i];
      double s = sigma[i];
      double e = exp(lambda * t);
      gsl_matrix_set (J, i, 0, e/s);
      gsl_matrix_set (J, i, 1, t * A * e/s);
      gsl_matrix_set (J, i, 2, 1/s);

    }
  return GSL_SUCCESS;
}

int expg_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  expg_f (x, params, f);
  expg_df (x, params, J);

  return GSL_SUCCESS;
}

int gauss_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double B = gsl_vector_get (x, 0);
  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
	  double diff=X[i]-C;
      double Yi = A*exp(-0.5*diff*diff/(w*w))+B;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int gauss_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /* Yi = y=A*exp[-(Xi-xc)^2/(2*w*w)]+B  */
      /* and the xj are the parameters (B,A,C,w) */
    
	  double s=sigma[i];
	  double diff=X[i]-C;
      double e =exp(-0.5*diff*diff/(w*w))/s;
	 
      gsl_matrix_set (J, i, 0, 1/s);
      gsl_matrix_set (J, i, 1, e);
      gsl_matrix_set (J, i, 2, diff*A*e/(w*w));
	  gsl_matrix_set (J, i, 3, diff*diff*A*e/(w*w*w));
    }
  return GSL_SUCCESS;
}

int gauss_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  gauss_f (x, params, f);
  gauss_df (x, params, J);

  return GSL_SUCCESS;
}

int lorentz_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double B = gsl_vector_get (x, 0);
  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
	  double diff=X[i]-C;
      double Yi = A*w/(4*diff*diff+w*w)+B;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int lorentz_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 1);
  double C = gsl_vector_get (x, 2);
  double w = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /* Yi = y=A*w/[4*(Xi-xc)^2+w^2]+B  */
      /* and the xj are the parameters (B,A,C,w) */
    
	  double s=sigma[i];
	  double diff=X[i]-C;
      double num =1/(4*diff*diff+w*w);
	  double den =4*diff*diff-w*w;
	 
      gsl_matrix_set (J, i, 0, 1/s);
      gsl_matrix_set (J, i, 1, w*num/s);
      gsl_matrix_set (J, i, 2, 8*diff*A*w*num*sqrt(num)/s);
	  gsl_matrix_set (J, i, 3, den*A*num*num/s);
    }
  return GSL_SUCCESS;
}

int lorentz_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  lorentz_f (x, params, f);
  lorentz_df (x, params, J);

  return GSL_SUCCESS;
}

int sin_f (const gsl_vector * x, void *params,
        gsl_vector * f)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *y = ((struct data *)params)->y;
  double *sigma = ((struct data *) params)->sigma;

  double B = gsl_vector_get (x, 0);
  double A = gsl_vector_get (x, 1);
  double T = gsl_vector_get (x, 2);
  double fi = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      double Yi = A*sin(8*atan(1.0)*X[i]/T+fi)+B;
      gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
    }

  return GSL_SUCCESS;
}

int sin_df (const gsl_vector * x, void *params,
         gsl_matrix * J)
{
  size_t n = ((struct data *)params)->n;
  double *X = ((struct data *)params)->X;
  double *sigma = ((struct data *) params)->sigma;

  double A = gsl_vector_get (x, 1);
  double T = gsl_vector_get (x, 2);
  double fi = gsl_vector_get (x, 3);

  size_t i;

  for (i = 0; i < n; i++)
    {
      /* Jacobian matrix J(i,j) = dfi / dxj, */
      /* where fi = (Yi - yi)/sigma[i],      */
      /* Yi = y=A*sin(2*pi*t/T+fi]+B  */
      /* and the xj are the parameters (B,A,T,fi) */
    
	  double s=sigma[i];
	  double S=sin(8*atan(1.0)*X[i]/T+fi);
      gsl_matrix_set (J, i, 0, 1/s);
      gsl_matrix_set (J, i, 1, S/s);
      gsl_matrix_set (J, i, 2, A*S*8*atan(1.0)*X[i]/(T*T*s));
	  gsl_matrix_set (J, i, 3, A*S/s);
    }
  return GSL_SUCCESS;
}

int sin_fdf (const gsl_vector * x, void *params,
          gsl_vector * f, gsl_matrix * J)
{
  sin_f (x, params, f);
  sin_df (x, params, J);

  return GSL_SUCCESS;
}


int print_state (size_t iter, gsl_multifit_fdfsolver * s)
{
FILE *stream;
stream=fopen("test.txt","a+");

fprintf (stream,"iter: %3u x = % 15.8f % 15.8f % 15.8f "
          "|f(x)| = %g\n",
          iter,
          gsl_vector_get (s->x, 0),
          gsl_vector_get (s->x, 1),
          gsl_vector_get (s->x, 2),
		  //gsl_vector_get (s->x, 3),
          gsl_blas_dnrm2 (s->f));
fclose(stream);
return 0;
}
