#ifndef GRAPH_H
#define GRAPH_H

#include <qwidget.h>
#include <qobject.h>
#include <qarray.h>
#include <qfont.h>
#include <qpen.h>

#include <qwt_plot.h>
#include <qwt_marker.h>
#include <qwt_plot_classes.h>

#include "curvesDialog.h"
#include "plotDialog.h"
#include "worksheet.h"
#include "genDialog.h"

class QPopupMenu;
class QStringList;
class QTabDialog;
class QRadioButton;
class QLineEdit;
class QComboBox;
class QMainWindow;
class QSpinBox;
class QwtPlotZoomer;
	
class Table;
class LegendMarker;
class LineMarker;
class ImageMarker;
class TitlePicker;
class ScalePicker;
class CanvasPicker;
class QwtPieCurve;
class Plot;

class Graph: public QWidget
{
	Q_OBJECT

public:
    Graph (QWidget* parent=0, const char* name=0, WFlags f=0);
	~Graph();

	enum AxisType{Numeric = 0, Txt = 1, Day = 2, Month = 3, Time = 4, Date = 5};

	enum MarkerType{None=-1, Text = 0, Arrow=1, Image=2};

	enum LabelFormat{Automatic = 0, Decimal = 1, Scientific = 2, Superscripts = 3};
	
	enum PlotType{Line = 0,
				  Scatter = 1, 
				  LineSymbols = 2,
				  VerticalBars = 3,
				  Area = 4,
				  Pie = 5,
				  VerticalDropLines = 6,
				  Spline = 7,
				  Steps = 8,
				  Histogram = 9,
				  HorizontalBars = 10,
				  VectXYXY = 11};

	Plot *myPlot;
	QwtPlotZoomer *d_zoomer;
	TitlePicker *titlePicker;
	ScalePicker *scalePicker;
	CanvasPicker* cp;
	 
public slots:
	Plot* plotWidget(){return myPlot;};

	// pie curves	
	bool isPiePlot();
	QwtPieCurve* pieCurve();
	void plotPie(QwtPieCurve* curve);
	void plotPie(Table* w,const QString& name);
	void plotPie(Table* w,const QString& name,const QPen& pen, int brush, int size, int firstColor);
	void updatePie(const QPen& pen, const Qt::BrushStyle &brushStyle, int size, int firstColor);
	void updatePieCurveData(Table* w, const QString& yColName, int curve);
	void removePie();
	QString pieLegendText();
	QPen pieCurvePen();
	Qt::BrushStyle pieBrushStyle();
	int pieFirstColor();
	int pieSize();
	QString savePieCurveLayout();

	bool plot(Table* w, const QString& name, int style);
	bool plot(Table* w, int xcol, const QString& name, int style);
	bool plot(Table* w, const QString& xColName, const QString& yColName, int style);

	void removeCurve(int index);
	void removeCurve(const QString& s);
	
	void updateData(Table* w, int curve);
	void updateCurveData(Table* w, const QString& yColName, int curve);
	 
	 int curves();
	 int curveDataSize(int curve);
	 double curveXValue(int curve,int point);
	 double minXValue(int curve);
	 double maxXValue(int curve);
	 double selectedXStartValue();
	 double selectedXEndValue();
	 long curveKey(int curve);
	 int curveIndex(long key);
	 QwtPlotCurve *curve(int index);
	 void insertPlottedList(const QStringList& names);
	 QStringList plotCurvesList();
	 QStringList plotAssociations();
	 void setPlotAssociations(const QStringList& newList);
	 void changePlotAssociation(Table* t, int curve, const QString& text);

	QStringList errorBarsAssociations();
	void setErrorBarsAssociations(const QStringList& newList);

	 void print();
	 void copyImage();
	 void exportGraph();
	 QRect exportToEPS(const QString& fname);
	 static void addBoundingBox(const QString& fname, const QRect& rect);
	 
	 void clearPlot();
	 void replot();
	 void updatePlot();

	 QPixmap graphPixmap();
	 void showExportDialog(const QString& fname, const QString& fileType);
	 void exportImage(const QString& fileName,const QString& fileType, int quality, bool transparent);

	  // error bars
	 void addErrorBars(Table *w, const QString& xColName, const QString& yColName, 
					   Table *errTable, const QString& errColName,
					   int type, int width, int cap, const QColor& color,
					   bool through, bool minus,bool plus);
	
	 void addErrorBars(Table *w, const QString& yColName, 
						 Table *errTable, const QString& errColName,
						 int type, int width, int cap, const QColor& color,
						 bool through, bool minus,bool plus);

	 void updateErrorBarsData(Table* w, int curve);
	 void updateErrorBars(int curve,bool xErr,int width,int cap,
		          const QColor& c,bool plus,bool minus,bool through);
				  
	 QStringList erorsLayout();

	 // event handlers 
	 void contextMenuEvent(QContextMenuEvent *);

	 // plot layout
	 void setPlotLimits(int axis,const QStringList& limits);
	 void setSecondaryPlotLimits(const QStringList& limits);
	 void setSecondaryPlotLimits(int axis,const QStringList& limits);
	 QStringList secondaryPlotLimits();
	 QStringList plotLimits();
	 void definePlotLimits(QStringList s);
	 QStringList xBottomAxisLimits();

	 static curveLayout initCurveLayout();
	 static curveLayout initFitCurveLayout();
	 QArray<curveLayout> allCurvesLayout();
	 void updateCurveLayout(int curve,const curveLayout& cL);

	 gridOptions getGridOptions();
	 void setGridOptions(const gridOptions& options);

	// zoom
	 void zoomed (const QwtDoubleRect &rect);
	 void zoom(bool on);
	 bool zoomOn();
	 
	 void movedPicker(const QPoint &pos);
	 void setAutoScale();
	 void updateScale();
	
	 // saving to file 
	 QString saveToString();
	 QString saveGridOptions();
	 QString saveScale();
	 QString saveScaleTitles();
	 QString saveFonts();
	 QString saveLegend();
	 QString saveMarkers();
	 QString saveCurveLayout(int index);
	 QString saveAxesTitleColors();
	 QString saveAxesColors();
	 QString saveEnabledAxes();
	 QString saveSecondaryPlotLimits();
	 QString saveErrorBars();
	 QString saveCanvasFrame();
	 QString saveTitle();
	 QString saveAxesTitleAlignement();
	 QString saveEnabledTickLabels();
	 QString saveTicksType();
	 QString saveCurves();
	 QString saveLabelsFormat();
	 QString saveLabelsRotation();
	 QString saveAxesLabelsType();
	 QString saveAxesBaseline();

	 // text markers 
	 void drawText(bool on);
	 bool drawTextActive();
	 
	 void insertTextMarker(LegendMarker* mrk);
	 long insertTextMarker(const QStringList& list);
	 long insertTextMarker_obsolete(const QStringList& list);
	 void updateTextMarker(const QString& text,int angle, int bkg,const QFont& fnt,
						   const QColor& textColor, const QColor& backgroundColor);
	
	 void removeMarker();
	 void cutMarker();
	 void copyMarker();
	 void pasteMarker();
	 void selectNextMarker();
	 void selectLineMarker(long markerID);
	 void selectTextMarker(long markerID);
	 void selectImageMarker(long markerID);
	 
	 MarkerType copiedMarkerType(){return selectedMarkerType;};
	 void setCopiedMarkerType(Graph::MarkerType type){selectedMarkerType=type;};
	 void setCopiedMarkerEnds(const QPoint& start, const QPoint& end);
 	 void setCopiedTextOptions(int bkg, const QString& text, const QFont& font, 
								const QColor& color, const QColor& bkgColor);
	 void setCopiedArrowOptions(int width, Qt::PenStyle style, const QColor& color,
													bool start, bool end, int headLength, int headAngle, bool filledHead);	
	 void setCopiedImageName(const QString& fn){auxMrkFileName=fn;};	
	 QRect copiedMarkerRect(){return QRect(auxMrkStart, auxMrkEnd);};
	 
	 //legendMarker
	 QwtArray<long> textMarkerKeys();
	 LegendMarker* textMarker(long id);
	 
	  // legend  
	 LegendMarker* legend();
	 void removeLegend();
	 void removeLegendItem(int index);
	 void addLegendItem(const QString& colName);
	 void insertLegend(const QStringList& lst);
	 void insertLegend_obsolete(const QStringList& lst);
	 void newLegend(const QFont& fnt, int frameStyle);
	 QSize newLegend(const QString& text);
	 bool legendOn();
	 
	 QString getLegendText();
	 void setLegendText(const QString& text);
	 
	 void setCurveLegend(const QString& text,int key);
	 QString legendText();//recreates the text from curves titles
	 
	 // line markers 
	 LineMarker* lineMarker(long id);
	  void insertLineMarker(LineMarker* mrk);
	 void insertLineMarker(QStringList list);
	 void updateLineMarker(const QColor& c,int w,Qt::PenStyle style,bool endArrow, bool startArrow);
	 QwtArray<long> lineMarkerKeys();
	 void drawLine(bool on);
	 bool drawLineActive();
	 void setArrowHeadGeometry(int length, int angle, bool filled);
	 void updateLineMarkerGeometry(const QPoint& sp,const QPoint& ep);

	 //image markers
	 ImageMarker* imageMarker(long id);
	 QwtArray<long> imageMarkerKeys();
	 void insertImageMarker(ImageMarker* mrk);
	 void insertImageMarker(const QPixmap& photo, const QString& fileName);
	 void insertImageMarker(const QStringList& options);
	 bool imageMarkerSelected();
	 void updateImageMarker(int x, int y, int width, int height);
	 
	 void resizeMarkers (double w_ratio, double h_ratio);
	 long selectedMarkerKey();
	 void setSelectedMarker(long mrk);
	  QwtPlotMarker* selectedMarkerPtr();
	 bool markerSelected();
	 bool arrowMarkerSelected();
	 void deselectItems();

	 // axes 
	 QValueList<int> axesType();
	 void setAxesType(const QValueList<int> tl); 
	 	
	 QStringList scalesTitles();
	 void setXAxisTitle(const QString& text);
	 void setYAxisTitle(const QString& text);
	 void setRightAxisTitle(const QString& text);
	 void setTopAxisTitle(const QString& text);
	 void setAxisTitle(int axis, const QString& text);

	 QFont axisTitleFont(int axis);
	 void setXAxisTitleFont(const QFont &fnt);
	 void setYAxisTitleFont(const QFont &fnt);
	 void setRightAxisTitleFont(const QFont &fnt);
	 void setTopAxisTitleFont(const QFont &fnt);
	 void setAxisTitleFont(int axis,const QFont &fnt);

	 void setAxisFont(int axis,const QFont &fnt);
	 QFont axisFont(int axis);
	 void initFonts(const QFont &scaleTitleFnt, const QFont &numbersFnt);
	
	 QColor axisTitleColor(int axis);
	 void setXAxisTitleColor(const QColor& c);
	 void setYAxisTitleColor(const QColor& c);
	 void setRightAxisTitleColor(const QColor& c);
	 void setTopAxisTitleColor(const QColor& c);
	 void setAxesTitleColor(QStringList l);
	 
	 int axisTitleAlignment (int axis);
	 void setXAxisTitleAlignment(int align);
	 void setYAxisTitleAlignment(int align);
	 void setTopAxisTitleAlignment(int align);
	 void setRightAxisTitleAlignment(int align);
	 void setAxesTitlesAlignment(const QStringList& align);

	 QStringList axesColors();
	 void setAxesColors(const QStringList& colors);
	 void showAxis(int axis, int type, const QString& formatInfo, Table *table, bool axisOn, int ticksType, 
					bool labelsOn, const QColor& c, int format, int prec, int rotation, int baselineDist);

	 QArray<bool> enabledAxes();
	 void enableAxes(QArray<bool> axesOn);
	 void enableAxes(const QStringList& list);	

	int labelsRotation(int axis);
	void setAxisLabelRotation(int axis, int rotation);
	
	QStringList enabledTickLabels();
	void setEnabledTickLabels(const QStringList& list);
	
	int axesLinewidth();
	void setAxesLinewidth(int width);
	void loadAxesLinewidth(int width);//used when opening a project file

	void drawAxesBackbones(bool yes);
	bool axesBackbones(){return drawAxesBackbone;};
	void loadAxesOptions(const QString& s);//used when opening a project file

	QValueList<int> axesBaseline();
	void setAxesBaseline(const QValueList<int> &lst);
	void setAxesBaseline(QStringList &lst);

	QValueList<int> ticksType();
	void setTicksType(const QValueList<int>& list);
	void setTicksType(const QStringList& list); 
	
	int minorTickLength();
	int majorTickLength();
	void setAxisTicksLength(int axis, int ticksType, int minLength, int majLength);
	void setTicksLength(int minLength, int majLength);
	void changeTicksLength(int minLength, int majLength);

	QStringList labelsNumericFormat();
	void setLabelsNumericFormat(const QStringList& list);
	void setLabelsNumericFormat(int axis, const QStringList& list);	
	void setLabelsNumericFormat(int axis, int format, int prec);
	void setLabelsDateTimeFormat(int axis, int type, const QString& formatInfo);
	
	QStringList axesLabelsFormatInfo(){return axesFormatInfo;};
	void setLabelsTextFormat(int axis, int type, const QString& labelsColName, Table *table);

	 // canvas frame 
	 void drawCanvasFrame(bool frameOn, int width);
	 void drawCanvasFrame(const QStringList& frame);
	 void drawCanvasFrame(bool frameOn, int width, const QColor& color);
	 QColor canvasFrameColor();
	 int canvasFrameWidth();
	 bool framed();

	  // plot title 
	 QString title();
	 void setTitle(const QString& t);
	 void setTitleFont(const QFont &fnt);
	 QFont titleFont();
	 QColor titleColor();
	 void setTitleColor(const QColor &c);
	 void removeTitle();
	 bool titleSelected();
	 void selectTitle();
	 void setTitleSelected(bool on);
	 int titleAlignment();
	 void setTitleAlignment(int align);
	 void initTitle( bool on);
	 void initTitleFont( const QFont& fnt);
	
	 // tools for modifing plot data 
	 void selectPoint(const QPoint &pos);
	 void highlightPoint(bool showIt);
	 void selectCurve(const QPoint &pos);
	 int selectedCurveID();
	 
	 void showCursor(bool showIt);
	 void shiftPointCursor(bool up);
	 void shiftCurveCursor(bool up);
	 void moveBy(int dx, int dy);
	 void move(const QPoint &pos);
	 void removePoint();
	 void movePoints(bool enabled);
	 bool movePointsActivated();
	 void removePoints(bool enabled);
	 bool removePointActivated();

	 void copyCanvas(bool on);
	 void enableCursor(bool on);
	 bool enabledCursor();
	 void showPlotPicker(bool on);
	 bool pickerActivated();
	
	// data range selectors
   	 bool selectorsEnabled();
	 bool enableRangeSelectors(bool on);
	 void disableRangeSelectors();
	 void moveRangeSelector(bool up);
	 void moveRangeSelector();
	 void shiftRangeSelector(bool shift);
	 void shiftCurveSelector(bool up);
	 int selectedPoints(long curveKey);

	 //border and margin
	 void changeMargin (int d);
	 void drawBorder (int width, const QColor& color);
	 void setBorder (int width, const QColor& color);
	 void setBackgroundColor(const QColor& color);

	 //functions in analysis.cpp file
	 QString integrateCurve(long curveKey,int order,int iter,double tol,double low,double up);
	 void diffCurve(long curveKey);
	 void fft(long curveKey, const QString & FFT);
	
	 QwtPlotCurve* getFitLimits(const QString& name, double from, double to,
								int params, int &start, int &end);

	 QString fitCurveLinear(long curveKey);

	 QString fitExpDecay(const QString& name, double damping, double amplitude, double yOffset);
	 QString fitExpDecay(const QString& name, double damping, double amplitude, double yOffset,
						   double from, double to, int iterations, int solver, double tolerance);
	 QString fitExpDecay(QwtPlotCurve *curve, double damping, double amplitude, double yOffset,
						   int start, int end, int iterations, int solver, double tolerance);
	 
	 QString fitExpDecay2(const QString& name, double firstTime, double secondTime,
						 double from, double yOffset);
	 QString fitExpDecay2(const QString& name, double amp1, double t1, double amp2, double t2, double yOffset,
						  double from, double to, int iterations, int solver, double tolerance);
	 QString fitExpDecay2(QwtPlotCurve *curve, double amp1, double t1, double amp2, double t2, double yOffset,
						   int start, int end, int iterations, int solver, double tolerance);

	 QString fitExpDecay3(const QString& name, double firstTime, double secondTime,
						double thirdTime, double from, double yOffset);
	 QString fitExpDecay3(const QString& name, double amp1, double t1, double amp2, double t2, 
						  double amp3, double t3, double yOffset, double from, double to, 
						  int iterations, int solver, double tolerance);
	 QString fitExpDecay3(QwtPlotCurve *curve, double amp1, double t1, double amp2, double t2, 
						  double amp3, double t3, double yOffset, int start, int end, 
						  int iterations, int solver, double tolerance);

	 QString fitExpGrowth(const QString& name, double damping, double amplitude, double yOffset);
	 QString fitExpGrowth(const QString& name, double damping, double amplitude, double yOffset,
						   double from, double to, int iterations, int solver, double tolerance);
	 QString fitExpGrowth(QwtPlotCurve *curve, double damping, double amplitude, double yOffset,
						   int start, int end, int iterations, int solver, double tolerance);

	 QString fitGauss(long curveKey);
	 QString fitGauss(const QString& name, double amplitude, double center, double width, double offset,
						   double from, double to, int iterations, int solver, double tolerance);
	 QString fitGauss(QwtPlotCurve *curve, double amplitude, double center, double width, double offset,
						   int start, int end, int iterations, int solver, double tolerance);

	 QString fitLorentz(long curveKey);
	 QString fitLorentz(const QString& name, double amplitude, double center, double width, double offset,
						   double from, double to, int iterations, int solver, double tolerance);
	 QString fitLorentz(QwtPlotCurve *curve, double amplitude, double center, double width, double offset,
						   int start, int end, int iterations, int solver, double tolerance);

	 QString fitPolynomial(const QString&,int order, int points,
				double start, double end, bool showFormula);
	 void setFitID(int id);
	 QString fitNonlinearCurve(const QString& curve,const QString& formula,
							const QStringList& params,const QStringList& paramsInit,
							double from,double to,int points, int solver, double tolerance);
							
	void interpolate(QwtPlotCurve *curve, int spline, int start, int end,
					int points, int colorIndex);


	//histograms
	void initHistogram(long curveID, const QArray<double>& Y, int it);
	void updateHistogram(Table* w, const QString& curveName, int curve);
	void updateHistogram(Table* w, const QString& curveName, int curve, bool automatic, 
							double binSize, double begin, double end);
	void updateBarsGap(int curve, int gapPercent);
	QString showHistogramStats(Table* w, const QString& curveName, int curve);

	//image analyse tools
	 bool lineProfile();
	 void calculateProfile(int average, bool ok);
	 void calculateLineProfile(const QPoint& start, const QPoint& end);
	 int averageImagePixel(const QImage& image, int px, int py, int average, bool moreHorizontal);
	 void showIntensityTable();
	 
	 //user defined functions
	 void addFunctionCurve(QString& type,QStringList &formulas,QStringList &vars,QValueList<double> &ranges,QValueList<int> &points);	 
	 //when reading from file
	 void insertFunctionCurve(const QString& formula, double from, double to, int points);
	 //for versions <0.4.3
	 void insertOldFunctionCurve(const QString& formula, double from, double step, int points);

	 void createWorksheet(const QString& name);
	 void activateGraph();
	 void moveGraph(const QPoint& pos);
	 void resizeGraph(const QPoint& pos);
	 void releaseGraph();
	 void newSizeGraph();

//vectXYXY curves
	void plotVectXYXY(Table* w, const QStringList& colList);
	void updateVectorsLayout(Table *w, int curve, int colorIndex, int width, 
								int arrowLength, int arrowAngle, bool filled, 
								const QString& xEndColName, const QString& yEndColName);
	void updateVectXYXY(Table* w,  int curve);
		
	//resizing			
	void setIgnoreResizeEvents(bool ok){ignoreResize=ok;};
	void resizeEvent ( QResizeEvent *e );
	void scaleFonts(double factor);

	void modified();

	static Qt::PenStyle getPenStyle(const QString& s);
	static Qt::PenStyle getPenStyle(int style);
	static Qt::BrushStyle getBrushStyle(int style);
	static QwtSymbol::Style getSymbolStyle(int type);
	static QColor color(int item);
	static void showPlotErrorMessage(QWidget *parent, const QStringList& emptyColumns);

signals:
    void releaseGraph(Graph*);
	void newSizeGraph(Graph*);
	void moveGraph(Graph*, const QPoint& pos);
	void resizeGraph(Graph*, const QPoint& pos);
    void selectedGraph (Graph*);
	void closedGraph(QWidget*);
	void drawTextOff();
	void drawLineOff();
	void cursorInfo(const QString&);
	void showPlotDialog();
	void showPieDialog();
	void createTable(const QString&,int,int,const QString&);
	void updateTable(const QString&,int,const QString&);
	void clearCell(const QString&,double);

	void viewImageDialog();
	void viewTextDialog();
	void viewLineDialog();
	void viewTitleDialog();
	void modifiedGraph();
	void modifiedGraph(Graph *);
	void hiddenPlot(QWidget*);

	void showContextMenu();
	void showMarkerPopupMenu();
	
	void axisDblClicked(int);	
	void xAxisTitleDblClicked();		
	void yAxisTitleDblClicked();
	void rightAxisTitleDblClicked();
	void topAxisTitleDblClicked();
	
	void createTablePlot(const QString&,int,int,const QString&);
	void createIntensityTable(const QPixmap&);
	void createHistogramTable(const QString&,int,int,const QString&);
	
private:
	QStringList axesFormatInfo;//stores columns used for axes with text labels or  time/date format info
	QValueList <int> axisType;
	QValueList <int> lblFormat; //stores label format used for the axes
	QwtDiMap xCanvasMap, yCanvasMap;
	gridOptions grid;
	MarkerType selectedMarkerType;
	QwtMarker::LineStyle mrklStyle;
	QStringList scales,associations,errorBarsLayout,errorBars;
	QArray<curveLayout> layout;
	QArray<long> keys;// curves on plot keys
	QArray<long> lines; // arrows on plot keys
	QArray<long> images; // images on plot keys
	QStringList tickLabelsOn;// tells wich axes have tick labels enabled
	QPen mrkLinePen;
	QFont auxMrkFont;
	QColor auxMrkColor, auxMrkBkgColor;
	QPoint auxMrkStart,auxMrkEnd;
	Qt::PenStyle auxMrkStyle;
	QString auxMrkFileName, auxMrkText;

	int selectedCurve, selectedPoint,startPoint,endPoint, selectedCursor, pieRay;
	int selectedCol,xCol,widthLine,cOnPlot,fitID,linesOnPlot;
	int auxMrkAngle,auxMrkBkg,auxMrkWidth, averagePixels;
	int auxArrowHeadLength, auxArrowHeadAngle;
	int axesLineWidth;
	long selectedMarker,legendMarkerID, startID, endID, functions;
	long mrkX,mrkY;//x=0 et y=0 line markers keys
	bool startArrowOn, endArrowOn, drawTextOn, drawLineOn;
	
	//the following bools tell iwhich data tool is activated by the user
	bool removePointsEnabled,movePointsEnabled;
	bool pickerEnabled, cursorEnabled, rangeSelectorsEnabled;
	
	bool piePlot;//tells if the plot is a pie plot
	bool lineProfileOn; // tells if pixel line profile is asked
	bool isTitleSelected, auxFilledArrowHead, ignoreResize;
	bool drawAxesBackbone;
};

class PrintFilter: public QwtPlotPrintFilter
{
public:
    PrintFilter(QwtPlot *plot) 
	{
	gridMajorColor=plot->gridMajPen().color();
	gridMinorColor=plot->gridMinPen().color();		
	};

	virtual QColor color(const QColor &c, Item item, int) const
    {
        if ( !(options() & PrintCanvasBackground))
        {
            switch(item)
            {
            case MajorGrid:
                return gridMajorColor;
            case MinorGrid:
                return gridMinorColor;
            default:	
                ;
            }
        }
        return c;
    }
	
	virtual QFont font(const QFont &f, Item, int) const
    {
        QFont f2 = f;
        f2.setPointSize(int(f.pointSize() * 1.5));
        return f2;
    }

private:
	QColor gridMajorColor,gridMinorColor;
};

#endif // GRAPH_H
